require 'stringio'

module UU
  module OS
    module HTTP

      # Module providing methods for multipart request parsing.
      module MultipartParser

        # Parses multipart body.
        # @param [String, IO] body Multipart request body
        # @param [String] boundary Multipart item boundary
        # @param [Symbol, String] binary_name Name of attribute containing large
        #    binary data. In case attribute with this name is found, its content
        #    is returned as stream and rest of multipart body is dumped (binary
        #    attribute should be always the last item of multipart body).
        # @return [Array<Hash<Symbol,Object>>] Parsed multipart request in for
        #   of +Array+ of +Hashes+. Each +Hash+ represents one form item with
        #   attributes +:name+ (item name), +:filename+ (optional), +:content_type+
        #   (optional) and +:body+ (+String+ or +IO+ based on +binary_name+ parameter).
        def parse_multipart(body, boundary, binary_name = nil)
          MultipartParser.parse_multipart(body, boundary, binary_name)
        end

        # @see #parse_multipart
        # @private
        def self.parse_multipart(body, boundary, binary_name = nil) # :nodoc:
          items = []
          binary_name = binary_name.to_s if !binary_name.nil?
          boundary = dequote(boundary)
          boundary_regex = /^--#{boundary}(--|\r\n)/
          body = MultipartBodyWrapper.new(body, boundary)          
          # Multipart item starts with boundary
          line = body.read_line
          while !line.nil? && line !~ boundary_regex do
            line = body.read_line
          end
          # Read multipart item
          line = body.read_line
          while !line.nil? do
            items << (item = {})
            # Read headers
            while !line.nil? && line !~ /^\r\n$/ do
              header = line.strip
              if header =~ /^content-type:(.*)/i
                item[:content_type] = $1.strip
              elsif header =~ /^content-disposition:(.*)/i
                value = $1.strip
                if value =~ /(^|;)[ ]*name=([^;]+)/i
                  item[:name] = dequote($2)
                end
                if value =~ /(^|;)[ ]*filename=([^;]+)/i
                  item[:filename] = dequote($2)
                end
              end
              line = body.read_line
            end
            # Read body. In case binary attribute is requested, it is
            # returned as stream and any following attributes are dumped
            # (because we cannot store binary data in memory or tempfiles).
            if !item[:name].nil? && binary_name == item[:name]
              item[:body] = body.read_binary
              return items
            else
              item[:body] = body.read_string
            end
            # Skip empty lines after item boundary to reach next header (or end of stream).
            line = body.read_line
            while !line.nil? && line =~ /^\r\n$/
              line = body.read_line
            end
          end
          return items
        end

        private

        # Removes optional quotes from given string.
        # @param [String] value Quoted string
        # @return [String] Dequoted string
        # @private
        def self.dequote(value) # :nodoc:
          value = value.strip
          value = value[1..-2] if value =~ /^".*"$/
          return value
        end
        private_class_method :dequote

        # Wrapper providing necessary methods for reading of body content.
        # @private
        class MultipartBodyWrapper # :nodoc:

          # Creates new instance of body wrapper.
          # @param [String, IO] body Multipart request body
          # @param [String] boundary Multipart boundary
          # @param [Fixnum] page_size Size of page (in bytes) to
          #   be read at once from body.
          def initialize(body, boundary, page_size = 32768)
            @body = body.respond_to?(:read) ? body : StringIO.new(body.to_s)
            @boundary = boundary
            @page_size = page_size
            @buffer = ''
          end

          # Reads one line (separated by CLRF) from body.
          # @return [String] Line
          def read_line
            line = ''
            tmp = "#{@buffer}#{@body.read(@page_size - @buffer.size)}"
            return nil if tmp.empty? || tmp.nil? #EOF
            while !tmp.empty? && tmp !~ /\r\n/
              line << tmp
              tmp = "#{@body.read(@page_size)}"
            end
            part, @buffer = tmp.split(/\r\n/, 2)
            if @buffer.nil?
              @buffer = ''
              line << "#{part}"
            else
              line << "#{part}\r\n"
            end
            return line.empty? ? nil : line
          end

          # Reads page (of size given in construtor) from body.
          # @return [String] Page
          def read_page
            page = "#{@buffer}#{@body.read(@page_size - @buffer.size)}"
            @buffer = ''
            return (page.empty? || page.nil?) ? nil : page
          end

          # Reads single parameter value (body is read until
          # boundary is reached). Value is returned as string.
          # @return [String] Parameter value
          def read_string
            boundary_regex = /^--#{@boundary}(--|\r\n)/
            value = ''
            prev_line = ''
            line = read_line
            while !line.nil? && !(prev_line =~ /\r\n$/ && line =~ boundary_regex) do
              value << prev_line
              prev_line, line = line, read_line
              tmp = 0
            end
            value << prev_line[0..-3]
            return value
          end

          # Reads single parameter value (body is read until
          # boundary is reached). Value is returned as IO.
          # Remaining body is dumped once this method is used.
          # @return [IO] Parameter value
          def read_binary
            boundary_regex = /\r\n--#{@boundary}(--|\r\n)/
            read, write = IO.pipe
            write.write('')
            Thread.new {
              prev_chunk = ''
              chunk = read_page
              # Read data until end of stream or boundary is reached
              while !chunk.nil? && (block = ('' << prev_chunk) << chunk) !~ boundary_regex do
                write.write(prev_chunk)
                prev_chunk, chunk = chunk, read_page
              end
              # Write last block (without boundary)
              write.write(block.split(boundary_regex)[0])
              # Dump rest of stream
              while !read_page.nil? do
                Thread.pass
              end
              write.close()
            }
            return read
          end

        end

      end

    end
  end
end
